--
-- $Id: Database.hs,v 1.13 2005/12/16 23:16:25 aamine Exp $
--
-- Copyright (c) 2005 Minero Aoki
--
-- This program is free software.
-- You can distribute/modify this program under the terms of
-- the GNU LGPL, Lesser General Public License version 2.1.
-- For details of the GNU LGPL, see the file "COPYING".
--

module Database
    (Database, fromConfig,
     pageSource, pageEncoding, savePageSource, doesPageExist,
     pageNames, pageNamesWithMtime) where

import Config
import URLEncoding
import FileUtils
import PathUtils
import Data.List
import Control.Monad
import System.IO
import System.IO.Error
import System.Posix.IO
import System.Posix.Unistd (sleep)
import System.Directory
import System.Time

data Database = Database { prefix :: String, encoding :: String }

fromConfig :: Config -> Database
fromConfig conf =
    Database { prefix = confLookupPath "directory" conf,
               encoding = confLookupString "encoding" conf }

pageSource :: Database -> String -> IO String
pageSource db name = readFile (pagePath db name)

pageEncoding :: Database -> String
pageEncoding db = encoding db

savePageSource :: Database -> String -> String -> IO ()
savePageSource (Database { prefix = dir }) name content =
    do let destdir = joinPath dir "pages"
           tmpdir = joinPath dir "tmp/pages"
       makePath destdir
       makePath tmpdir
       atomicWriteFile destdir tmpdir (encodeName name) content

doesPageExist :: Database -> String -> IO Bool
doesPageExist db name = doesFileExist (pagePath db name)

pageNames :: Database -> IO [String]
pageNames db = return . map (decodeName) =<< fileEntries (prefix db ++ "/pages")

pageNamesWithMtime :: Database -> IO [(String, CalendarTime)]
pageNamesWithMtime db =
    do names <- pageNames db
       return . zip names =<< mapM (mtime . pagePath db) names

mtime path = toCalendarTime =<< getModificationTime path

pagePath db name = concatPath [prefix db, "pages", encodeName name]

encodeName = urlencode
decodeName = urldecode

atomicWriteFile destdir tmpdir name content = atomicWriteFile' 5
  where
    atomicWriteFile' 0 = ioError (userError "failed to lock file")
    atomicWriteFile' nRetry =
        do catch (exclWriteFile (joinPath tmpdir name) content)
                 (\err -> if isAlreadyExistsError err then
                              do sleep 1
                                 atomicWriteFile' (nRetry - 1)
                          else
                              ioError err)
           catch (renameFile (joinPath tmpdir name) (joinPath destdir name))
                 (\err -> do forceRemove (joinPath tmpdir name)
                             ioError err)

    exclWriteFile path content = do h <- fdToHandle =<< exclCreate path
                                    hPutStr h content
                                    hClose h

    exclCreate path = openFd path WriteOnly (Just 0o666)
                          (defaultFileFlags { exclusive = True })
