--
-- $Id: Main.hs,v 1.21 2005/12/16 20:44:54 aamine Exp $
--
-- Copyright (c) 2005 Minero Aoki
--
-- This program is free software.
-- You can distribute/modify this program under the terms of
-- the GNU LGPL, Lesser General Public License version 2.1.
-- For details of the GNU LGPL, see the file "COPYING".
--

module Main (main) where

import Config
import SimpleCGI
import Database
import Syntax
import ParameterizedText
import Template
import URLMapper
import URLEncoding
import HTMLUtils
import Data.List
import Data.Maybe
import System.Time
import System.Locale (defaultTimeLocale)

frontPageName = "FrontPage"

main = do ctx <- loadContext "./config"
          webApplication (wikiSession ctx . wikiRequest)

data Context = Context Database TemplateRepository URLMapper

loadContext :: String -> IO Context
loadContext path =
    do conf <- loadConfig path
       return $ Context (Database.fromConfig (confLookup "database" conf))
                        (Template.fromConfig (confLookup "template" conf))
                        (URLMapper.fromConfig (confLookup "urlmapper" conf))

data WikiRequest = ViewRequest String
                 | EditRequest String
                 | SaveRequest String String
                 | RecentRequest

wikiRequest :: HTTPRequest -> WikiRequest
wikiRequest req =
    case (lookupVar "cmd" req, lookupVar "name" req) of
      (Nothing,     Just name) -> ViewRequest name
      (Just "view", Just name) -> ViewRequest name
      (Just "edit", Just name) -> EditRequest name
      (Just "save", Just name) -> case lookupVar "text" req of
                                    Just cs -> SaveRequest name cs
                                    Nothing -> ViewRequest frontPageName
      (Just "recent", _)       -> RecentRequest
      _ -> ViewRequest frontPageName

wikiSession :: Context -> WikiRequest -> IO HTTPResponse
wikiSession (Context db tmpl umap) req =
    catch (respondTo req) (const $ frontPageResponse)
  where
    respondTo (ViewRequest name) =
        catch (viewPageResponse name) (const $ editPageResponse name)

    respondTo (EditRequest name) = editPageResponse name

    respondTo (SaveRequest name content) =
        do savePageSource db name content
           return $ softRedirectResponse (pageURL umap name)

    respondTo (RecentRequest) = recentResponse

    frontPageResponse = viewPageResponse frontPageName

    viewPageResponse name =
        return . HTTPResponse pageContentType
        =<< fill "view" name =<< pageHTML db name

    editPageResponse name =
        return . HTTPResponse pageContentType
        =<< fill "edit" name =<< catch (pageSource db name) (\ex -> return "")

    recentResponse =
        do body <- (return . makeRecentPage . sortBy (newerFirst)
                    =<< pageNamesWithMtime db)
           html <- fillTemplate tmpl "recent"
               [ ("cgiURL", escape (cgiURL umap)),
                 ("encoding", pageEncoding db),
                 ("content", body) ]
           return (HTTPResponse pageContentType html)

    newerFirst (_,a) (_,b) = b `compare` a

    makeRecentPage = ol . unlines . map (makeRecentPageCol)

    makeRecentPageCol (name, mtime) =
        li (formatTime mtime ++ ": " ++ pageAnchor umap name)

    pageAnchor umap name = a_href (pageURL umap name) name

    pageContentType = textContentType "text/html" $ pageEncoding db

    pageHTML db name =
        do items <- (return . Syntax.compile =<< pageSource db name)
           emap <- makeExistenceMap . map (hrefName) . parameters $ items
           return (ParameterizedText.resolve (resolveHref emap) items)

    makeExistenceMap = mapM (\n -> return . (,) n =<< doesPageExist db n)

    hrefName (PageLink name) = name
    -- hrefName _ = error "hrefName: arg is not PageLink"

    resolveHref emap (PageLink name) =
        if fromJust (lookup name emap)
        then a_href (pageURL umap name) name
        else a_href (pageEditURL umap name) "?" ++ name

    fill id name content = fillTemplate tmpl id
        [ ("cgiURL", escape (cgiURL umap)),
          ("encoding", pageEncoding db),
          ("pageName", escape name),
          ("encodedPageName", escape (urlencode name)),
          ("content", content) ]

    softRedirectResponse url = HTTPResponse "text/html" $ concat $
        [ "<html>\n",
          "<head>\n",
          "<meta http-equiv=\"Refresh\" content=\"1;url=", escape url, "\">\n",
          "<title>Redirect</title>\n",
          "</head>\n",
          "<body>\n",
          "<p><a href=\"", escape url, "\">Redirect: ", escape url,"</a></p>\n",
          "</body>\n",
          "</html>\n" ]

formatTime = strftime "%Y-%m-%d %H:%M:%S %Z"

strftime fmt = formatCalendarTime defaultTimeLocale fmt

ul s = "<ul>\n" ++ s ++ "</ul>"
ol s = "<ol>\n" ++ s ++ "</ol>"
li s = "<li>" ++ s ++ "</li>"
