--
-- $Id: Syntax.hs,v 1.14 2005/12/16 22:14:51 aamine Exp $
--
-- Copyright (c) 2005 Minero Aoki
--
-- This program is free software.
-- You can distribute/modify this program under the terms of
-- the GNU LGPL, Lesser General Public License version 2.1.
-- For details of the GNU LGPL, see the file "COPYING".
--

module Syntax (HTML, Href(..), compile) where

import HTMLUtils
import ParameterizedText
import TextUtils
import Text.ParserCombinators.Parsec
import Data.List
import Data.Char

type HTML = ParameterizedText Href

data Href = PageLink { hrefPageName :: String }

compile :: String -> HTML
compile = compileDocument . lines

compileDocument :: [String] -> HTML
compileDocument []        = []
compileDocument ("":cont) = compileDocument cont
compileDocument doc@(line@(c:_):cont)
    | isBlank line = compileDocument cont
    | c == '='     = heading line ++ compileDocument cont
    | c == '*'     = consume (takeUL) (ulist) doc
    | c == '#'     = consume (takeOL) (olist) doc
    | c == ':'     = consume (takeDL) (dlist) doc
    | isSpace c    = consume (takePre) (preformatted) doc
    | otherwise    = consume (takeP) (paragraph) doc
  where
    consume fetch compile xs = compile block ++ compileDocument rest
                               where (block, rest) = fetch xs

--
-- Takers
--

takeUL  = takeMarkedBlock '*'
takeOL  = takeMarkedBlock '#'
takeDL  = takeMarkedBlock ':'
takePre = span (beginWith isSpace)
takeP   = span (beginWith isNoFunc)

takeMarkedBlock mark = span (\line -> isMarked line || listContinued line)
  where
    isMarked = ([mark] `isPrefixOf`)

listContinued = isIndented

isIndented = beginWith (isSpace)

beginWith f ""    = False
beginWith f (c:_) = f c

isNoFunc c = not (c `elem` "=*#: \t\r\n\v\f")

--
-- Compilers
--

heading :: String -> HTML
heading line = let (mark, label) = span (== '=') line
               in h (length mark) [Text (escape $ strip label)]

ulist :: [String] -> HTML
ulist = compileList '*' ul

olist :: [String] -> HTML
olist = compileList '#' ol

data ListItem = Leaf String | Node [ListItem]

compileList :: Char -> (HTML -> HTML) -> [String] -> HTML
compileList mark markup =
    markup . concatMap (compileItem) . mktree . levelize . unwrap
  where
    compileItem :: ListItem -> HTML
    compileItem (Leaf str)   = li . compileText $ str
    compileItem (Node items) = li . markup . concatMap (compileItem) $ items

    mktree :: [(Int, String)] -> [ListItem]
    mktree = mktreeLev 1
      where
        mktreeLev _ [] = []
        mktreeLev n (x:xs)
            | fst x == n  = Leaf (snd x) : mktreeLev n xs
            | otherwise   = let (g,gs) = span ((> n) . fst) (x:xs)
                            in Node (mktreeLev (n+1) g) : mktreeLev n gs

    levelize :: [String] -> [(Int, String)]
    levelize [] = []
    levelize (x:xs) = (length marks, strip content) : levelize xs
                      where (marks, content) = span (==mark) x

    unwrap :: [String] -> [String]
    unwrap = reverse . unwrap' . reverse
      where
        unwrap' [] = []
        unwrap' [x] = [x]
        unwrap' (x:x2:xs)
            | listContinued x = unwrap' ((x2 ++ "\n" ++ strip x) : xs)
            | otherwise       = x : unwrap' (x2:xs)

dlist :: [String] -> HTML
dlist = dl . concatMap (parseLine)
  where
    parseLine line =
        case break (==':') (tail line) of
          (t,(':':d)) -> (dt.compileText.strip $ t)++(dd.compileText.strip $ d)
          (t,      _) -> (dt.compileText.strip $ t)

preformatted :: [String] -> HTML
preformatted xs = pre . (:[]) . Text . escape
        . concat . intersperse "\n"
        . map (unindent $ minimumIndent xs) $ xs

unindent :: Int -> String -> String
unindent n line = let (sp, str) = span (isSpace) . detab 8 $ line
                  in drop n sp ++ str

minimumIndent :: [String] -> Int
minimumIndent = minimum . map (numIndent)

numIndent :: String -> Int
numIndent = length . fst . span (isSpace) . detab 8

detab :: Int -> String -> String
detab _ []        = ""
detab ts ('\t':cs) = replicate ts ' ' ++ detab ts cs
detab ts (c:cs)    = c : detab ts cs

paragraph :: [String] -> HTML
paragraph = p . compileText . concat . intersperse "\n"

compileText :: String -> HTML
compileText str = case (parse text "" str) of
                    Right partext -> partext
                    Left err -> p [Text (escape $ show err)]

text = many component
  where
    component =  do s <- many1 (noneOf ('h' : charsBetween 'A' 'Z'))
                    return (Text (escape s))
             <|> do s <- try(wikiName)
                    return (Param $ PageLink s)
             <|> do u <- try(urlAutoLink)
                    return (Text $ a_href u u)
             <|> do c <- anyChar
                    return (Text $ escapeChar c)

charsBetween beg end =
    map (chr) . snd . splitAt (ord beg) . take (ord end + 1) $ [0..]

wikiName = moreThan 2 word >>= (return.concat)
  where
    word = do c <- upper
              s <- many1 (lower <|> digit)
              return (c:s)

urlAutoLink = do a <- string "http"
                 b <- option "" (string "s")
                 c <- string "://"
                 d <- many1 urlChar
                 return $ concat [a,b,c,d]

urlChar = alphaNum <|> oneOf ";/?:@&=+$,-_.!~*'#%"  -- '(' ')'

moreThan n p = do xs <- count n p
                  ys <- many p
                  return (xs ++ ys)

h :: Int -> HTML -> HTML
h n = joinWithTag ('h' : show level) False
      where level = maximum [1, minimum [6, n]]

ul, ol, li, dl, dt, dd, pre, p :: HTML -> HTML
ul = joinWithTag "ul" True
ol = joinWithTag "ol" True
li = joinWithTag "li" False
dl = joinWithTag "dl" True
dt = joinWithTag "dt" False
dd = joinWithTag "dd" False
pre = joinWithTag "pre" False
p  = joinWithTag "p" False

joinWithTag :: String -> Bool -> HTML -> HTML
joinWithTag name addEOL content =
    Text ("<"++name++">" ++ eol) : (content ++ [Text ("</"++name++">\n")])
  where
    eol = if addEOL then "\n" else ""
